## 
## @package locstats
## @author Dimitri Kokkonis ([\@kokkonisd](https://github.com/kokkonisd))
## 
## This file contains useful definitions, such as the BASE_DIR path, and also 
## some helper functions.
##

import os
import json

VERSION = "1.0.9"

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
LANG_FILE = os.path.join(BASE_DIR, "languages.json")
with open(LANG_FILE, "r") as ext:
    LANG_DATA = json.loads(ext.read())


def info(message, ending = '\n'):
    """Prints a colored info message."""
    print(f"\033[33m{message}\033[0m", end = ending)


def warn(message, ending = '\n'):
    """Prints a colored warning message."""
    print(f"\033[91m/!\\ {message}\033[0m", end = ending)


def fail(message, ending = '\n'):
    """Prints a colored error message."""
    print(f"\033[31m{message}\033[0m", end = ending)


def esc_regex(string):
    """Escapes non-literal regex characters in a string."""
    return string.replace('[', '\\[')\
                 .replace(']', '\\]')\
                 .replace('\\', '\\\\')\
                 .replace('^', '\\^')\
                 .replace('$', '\\$')\
                 .replace('.', '\\.')\
                 .replace('|', '\\|')\
                 .replace('?', '\\?')\
                 .replace('*', '\\*')\
                 .replace('+', '\\+')\
                 .replace('(', '\\(')\
                 .replace(')', '\\)')



## 
## @package locstats
## @author Dimitri Kokkonis ([\@kokkonisd](https://github.com/kokkonisd))
## 
## This is the entry point for the `locstats` tool.
## 

import sys
import os
import click

from .definitions import VERSION, LANG_DATA, info, fail
from .loc import get_source_files, get_loc


@click.command()
@click.argument('language', nargs = 1) # Exactly one argument
@click.argument('src_dirs', nargs = -1) # Unlimited arguments
@click.option('--strict',
              is_flag = True,
              default = False,
              help = "Run in strict mode (ignore comments and empty lines).")
@click.option('-m', '--minimal',
              is_flag = True,
              default = False,
              help = "Give minimal output (just the LOC count).")
@click.option('--silent',
              is_flag = True,
              default = False,
              help = "Silence all warnings (such as directories not being "\
                     "found).")
@click.option('-d', '--detailed',
              is_flag = True,
              default = False,
              help = "Output a detaled list of LOC per file.")
@click.version_option(version = VERSION,
                      prog_name = "locstats")
def main(language, src_dirs, strict, minimal, silent, detailed):
    """Counts the LOC in a given language in a given directory set."""

    # Convert user-fed language to lowercase (see `languages.json`)
    language = language.lower()
    # Check if language exists in database
    if language not in LANG_DATA:
        info(f"The language `{language}` doesn't exist or hasn't yet been "\
              "registered in our database.")

        info("\nHere's a list of all the languages we currently support:")
        print(f"{', '.join(sorted(list(LANG_DATA.keys())))}\n")

        info("If you'd like to contribute, you can check out locstats' "\
             "GitHub page: https://github.com/kokkonisd/locstats")
        exit(1)


    loc_count_per_file = []

    for src in src_dirs:
        # Get all the source files from the given directories
        source_files = get_source_files(
            src_dir = src,
            src_extensions = LANG_DATA[language]["extensions"],
            silent = silent
        )

        for filename in source_files:
            # Count the LOC in each file
            loc_count_per_file.append((
                filename,
                get_loc(filename = filename,
                        strict = strict,
                        comments = LANG_DATA[language]["comments"],
                        silent = silent)
            ))


    # Remove files that don't count towards LOC
    loc_count_per_file = list(filter(lambda x: x[1][0] > 0, loc_count_per_file))

    total_loc_count = sum(x[1][0] for x in loc_count_per_file)
    comm_line_count = sum(x[1][1] for x in loc_count_per_file)

    # Give the LOC count to the user
    if minimal:
        # Just print the number
        print(total_loc_count)
    elif detailed:
        # Print the filenames along with their LOC count and the percentage of
        # the total LOC count they represent
        
        # Sort the files by descending LOC count
        loc_count_per_file = sorted(loc_count_per_file,
                                    key = lambda x: x[1][0],
                                    reverse = True)

        # Get the maximum filename length (in order to format output later)
        max_filename_length = max(list(map(lambda x: len(x[0]),
                                           loc_count_per_file)))

        # Print the "table"'s header
        print(f"FILENAME {' ' * (max_filename_length - 8)} LOC (%)")
        print("-" * (max_filename_length + 20))

        # Print the files, their LOC count and their % respective to the total
        # LOC count
        for filename, (loc, comm_lines) in loc_count_per_file:
          print(f"{filename}{' ' * (max_filename_length - len(filename))}: {loc} ({loc / total_loc_count * 100:.2f}%)")

        print("-" * (max_filename_length + 20))
        print(f"TOTAL LOC{' ' * (max_filename_length - 9)}: {total_loc_count} (100%)")
    else:
        print(f"You have written approximately {total_loc_count} LOC in {LANG_DATA[language]['official_name']}",
              end='')

        if (not strict) and (total_loc_count > 0):
            print(f", {comm_line_count / total_loc_count * 100:.2f}% of which are comments.")
        else:
            print(".")


if __name__ == "__main__":
    main()



import setuptools

from locstats.definitions import VERSION

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "locstats",
    version = VERSION,
    author = "Dimitri Kokkonis",
    author_email = "kokkonisd@gmail.com",
    description = "A statistics tool for your LOC per language",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/kokkonisd/locstats",
    packages = setuptools.find_packages(),
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires = [ "click" ],
    package_data = {'locstats': ['languages.json']},
    include_package_data = True,
    entry_points = {'console_scripts': [
        'locstats = locstats.__main__:main',
    ], },
)



## 
## @package locstats
## @author Dimitri Kokkonis ([\@kokkonisd](https://github.com/kokkonisd))
## 
## This file contains useful functions to get the number of LOC from a given
## directory set.
##

import os
import re

from .definitions import esc_regex, warn


def get_source_files(src_dir, src_extensions, silent):
    """Returns a list of source files given a root directory and a file extension."""
    source_files = []

    if not os.path.exists(src_dir):
        if not silent:
            warn(f"Source directory `{src_dir}` doesn't exist. Skipping.")
        return source_files

    if os.path.isfile(src_dir):
        if not silent:
            warn(f"`{src_dir}` is a file, not a directory. Skipping.")
        return source_files


    for (dirpath, dirnames, filenames) in os.walk(src_dir):
        for file in filenames:
            name, extension = os.path.splitext(os.path.join(dirpath, file))
            if extension in src_extensions:
                source_files += [os.path.join(dirpath, file)]

    return source_files


def get_loc(filename, strict, comments, silent):
    """Returns the LOC count and the comment-line count given a file. Optionally strips out comments and blank
    lines."""
    with open(filename, "r") as source:
        try:
            lines = source.read()
        except:
            if not silent:
                warn(f"Could not read file `{filename}` (probably because it's not UTF-8). Skipping.")
            return (0, 0)
   
    comm_lines = []

    for comment in comments["single_line"]:
        # Simulate ^ and $ characters, as for some reason the re.MULTILINE flag doesn't work (and thus ^ and $ only
        # match at the beginning and at the end of the string, not at every line)
        comm_single_regex = f"((^|\n)[ \t]*{esc_regex(comment)}.+[ \t]*)"                 

        # If in strict mode, remove all single line comments
        if strict:
            lines = re.sub(comm_single_regex, "", lines)
        else:
            # Collect all single line comments
            comm_lines += list(map(lambda x: x[0], re.findall(comm_single_regex, lines)))


    for start, stop in comments["multi_line"]:
        comm_multi_regex = f"((^|\n)[ \t]*{esc_regex(start)}"\
                           f"((?!{esc_regex(stop)})[\\s\\S])*"\
                           f"{esc_regex(stop)}[ \t]*)" 


        # If in strict mode, remove all multi line comments
        if strict:
            lines = re.sub(comm_multi_regex, "", lines)
        else:
            # Collect all multi line comments
            raw_matches = list(map(lambda x: x[0], re.findall(comm_multi_regex, lines)))
            multi_comm_lines = list(map(lambda x: x.split('\n'), raw_matches)) 
            comm_lines += [item for sublist in multi_comm_lines for item in sublist]


    if strict:
        lines = list(filter(lambda x: len(x) > 0, lines.split('\n')))
    else:
        lines = lines.split('\n')
        # Clean up comm lines
        comm_lines = list(filter(lambda x: len(x) > 0, comm_lines))


    return len(lines), len(comm_lines)
